﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Interop;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using Windows.Data.Pdf;
using Windows.Graphics.Imaging;
using Windows.Media.Ocr;
using Windows.Storage.Streams;
using MS.Internal;
using BitmapEncoder = System.Windows.Media.Imaging.BitmapEncoder;
using BitmapFrame = System.Windows.Media.Imaging.BitmapFrame;
using Cursors = System.Windows.Input.Cursors;
using DataFormats = System.Windows.DataFormats;
using DragEventArgs = System.Windows.DragEventArgs;
using Image = System.Windows.Controls.Image;
using OpenFileDialog = Microsoft.Win32.OpenFileDialog;

namespace PDF2PNGUI
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        [System.Runtime.InteropServices.DllImport("gdi32.dll")]
        private static extern bool DeleteObject(IntPtr hObject);

        /// <summary>画像ファイルとして受け入れる拡張子</summary>
        private static readonly string[] IMAGE_EXTENTIONS = new string[] { ".png", ".bmp", ".jpg", ".jpeg", ".gif", ".tif", ".tiff" };

        /// <summary>変換元PDFのパス</summary>
        private string _pdfPath;
        /// <summary>変換後の画像ファイルの保存先フォルダパス</summary>
        private string _saveFolderPath;

        /// <summary>画像ファイルの保存先フォルダから読み込んだ画像を格納したImage</summary>
        private readonly List<Image> _images;

        private OcrEngine engine;


        /// <summary>
        /// 
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            this._images = new List<Image>();
        }

        #region ハンドラ

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_OnSizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (!IsInitialized) return;

            UpdateImages();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_OnDrop(object sender, DragEventArgs e)
        {
            // ドロップされたファイル名を取得
            var files = e.Data.GetData(DataFormats.FileDrop) as string[];
            if (files == null || files.Length == 0)
            {
                return;
            }
            TxtPdfPath.Text = files[0];
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_OnPreviewDragOver(object sender, DragEventArgs e)
        {
            // PDFだけを受け付けるようにする
            e.Handled = true;

            if (!e.Data.GetDataPresent(DataFormats.FileDrop, true))
            {
                e.Effects = System.Windows.DragDropEffects.None;
                return;
            }

            var files = e.Data.GetData(DataFormats.FileDrop) as string[];
            if (files.Length != 1)
            {
                e.Effects = System.Windows.DragDropEffects.None;
                return;
            }

            e.Effects = (System.IO.Path.GetExtension(files[0]).ToLower() == ".pdf") ? System.Windows.DragDropEffects.Copy : System.Windows.DragDropEffects.None;
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BtnPdfPath_OnClick(object sender, RoutedEventArgs e)
        {

            var result = OpenFileDialog("PDF ファイル(*.pdf)|*.pdf",
                "変換元PDFの選択",
                (String.IsNullOrEmpty(TxtPdfPath.Text)) ? "" : TxtPdfPath.Text);
            TxtPdfPath.Text = result.Item1 ? result.Item2 : "";
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BtnSaveFolder_OnClick(object sender, RoutedEventArgs e)
        {
            var result = OpenFolderDialog("変換後の画像ファイル保存先フォルダの選択",
                (String.IsNullOrEmpty(TxtSaveFolder.Text)) ? "" : TxtSaveFolder.Text);
            TxtSaveFolder.Text = result.Item1 ? result.Item2 : "";
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private async void BtnPerform_OnClick(object sender, RoutedEventArgs e)
        {
            // 変換元PDF
            _pdfPath = TxtPdfPath.Text.Trim();
            if (string.IsNullOrEmpty(_pdfPath))
            {
                System.Windows.MessageBox.Show("変換元PDFを指定してください", "エラー",
                    MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                BtnPdfPath.Focus();
                return;
            }

            if (!System.IO.File.Exists(_pdfPath))
            {
                System.Windows.MessageBox.Show("変換元PDFが存在しません", "エラー",
                    MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                BtnPdfPath.Focus();
                return;
            }

            // 変換後画像ファイル保存先
            if (ChkSaveFolder.IsChecked == true)
            {
                _saveFolderPath = TxtSaveFolder.Text.Trim();
                if (string.IsNullOrEmpty(_saveFolderPath))
                {
                    System.Windows.MessageBox.Show("変換後画像ファイル保存先を指定してください", "エラー",
                        MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                    BtnSaveFolder.Focus();
                    return;
                }

                if (!System.IO.Directory.Exists(_saveFolderPath))
                {
                    if (!System.IO.Directory.CreateDirectory(_saveFolderPath).Exists)
                    {
                        System.Windows.MessageBox.Show("変換後画像ファイル保存先の作成に失敗しました", "エラー",
                            MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                        BtnSaveFolder.Focus();
                        return;
                    }
                }
            }
            else
            {
                // 未指定の場合はデスクトップにPDFファイル名+タイムスタンプでフォルダを作成
                _saveFolderPath = System.IO.Path.Combine(System.Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory),
                        $"{System.IO.Path.GetFileNameWithoutExtension(_pdfPath)}_{DateTime.Now:yyyyMMddHHmmss}");
                if (!System.IO.Directory.CreateDirectory(_saveFolderPath).Exists)
                {
                    System.Windows.MessageBox.Show("変換後画像ファイル保存先の作成に失敗しました", "エラー",
                        MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                    BtnSaveFolder.Focus();
                    return;
                }
            }

            // DPI
            int dpi;
            if (!int.TryParse(TxtDpi.Text.Trim(), out dpi))
            {
                System.Windows.MessageBox.Show("DPIは数値で指定してください", "エラー",
                    MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                TxtDpi.Focus();
                return;
            }

            if (dpi < 16)
            {
                System.Windows.MessageBox.Show("DPIは16以上の数値で指定してください", "エラー",
                    MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
                TxtDpi.Focus();
                return;
            }

            // 画像形式
            // ラジオボタンのTagプロパティから拡張子を取得し保存処理に使う
            var selected = from x in SpnlFormat.Children.OfType<System.Windows.Controls.RadioButton>()
                           where x.IsChecked == true
                           select x;
            string extension = selected.FirstOrDefault()?.Tag as string ?? "png";

            try
            {
                this.IsEnabled = false;
                this.Cursor = Cursors.Wait;

                // ロード済みの画像を解放
                ClearImages();

                // PDFを画像に変換
                await Convert(_pdfPath, _saveFolderPath, dpi, extension, ChkOcr.IsChecked == true);

                // 変換した画像を読み込み表示
                LoadImages(_saveFolderPath);
                UpdateImages();

                // 変換後画像ファイルの保存先フォルダを開く
                System.Diagnostics.Process.Start("explorer.exe", _saveFolderPath);

                this.IsEnabled = true;
                this.Cursor = Cursors.Arrow;
            }
            catch (Exception ex)
            {
                this.IsEnabled = true;
                this.Cursor = Cursors.Arrow;

                System.Windows.MessageBox.Show($"PDFの画像変換処理中にエラーが発生しました\n({ex.Message})", "エラー",
                    MessageBoxButton.OK, MessageBoxImage.Error, MessageBoxResult.OK);
            }

        }

        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pdfpath"></param>
        /// <param name="saveFolderPath"></param>
        /// <param name="dpi"></param>
        /// <param name="extension"></param>
        /// <param name="ocrRecognize"></para>
        /// <returns></returns>
        private async Task<int> Convert(string pdfpath, string saveFolderPath, int dpi = 300, string extension = "png", bool ocrRecognize = true)
        {
            // PNGファイルの基本ファイル名を決定
            var basename = System.IO.Path.GetFileNameWithoutExtension(pdfpath);

            // PDFを読み込み、ページごとに画像に変換
            using (var instream = new MemoryStream(System.IO.File.ReadAllBytes(pdfpath)))
            using (var inrastream = instream.AsRandomAccessStream())
            {
                // 変数pdfdocは処理のあいだ開いていないとビットマップへのレンダリングに失敗する
                var pdfdoc = await PdfDocument.LoadFromStreamAsync(inrastream);

                // テキストファイル名を決定
                var textFilePath = System.IO.Path.Combine(saveFolderPath, $"{basename}_{DateTime.Now:yyyyMMddHHmmss}.txt");

                for (uint pageIndex = 0; pageIndex < pdfdoc.PageCount; pageIndex++)
                {
                    // 画像ファイル名を決定
                    var imageFilePath = System.IO.Path.Combine(saveFolderPath, $"{basename}_{pageIndex + 1:0000}.{extension}");

                    using (var pdfpage = pdfdoc.GetPage(pageIndex))
                    using (var outstream = new MemoryStream())
                    using (var outratream = outstream.AsRandomAccessStream())
                    {
                        // ページの内容をBitmapにレンダリング
                        var renderOptions = new Windows.Data.Pdf.PdfPageRenderOptions();
                        // 解像度などを指定のものに変換
                        renderOptions.DestinationWidth = (uint)((pdfpage.Size.Width / (96d / 72d)) * (dpi / 72d));
                        renderOptions.DestinationHeight = (uint)((pdfpage.Size.Height / (96d / 72d)) * (dpi / 72d));
                        await pdfpage.RenderToStreamAsync(outratream, renderOptions);

                        using (var image = System.Drawing.Image.FromStream(outstream))
                        {
                            // Bitmapの内容をPNGに変換
                            var bmpsource = ConvertToBitmapSource(image as Bitmap);
                            SaveImage(bmpsource, imageFilePath, extension);

                            // Bitmapの画像をOCRにかけてテキストを書き出し
                            if (ocrRecognize)
                            {
                                if (engine == null) engine = OcrEngine.TryCreateFromUserProfileLanguages();
                                var text = await RecognizeFromBitmapAsync(engine, bmpsource);
                                using (var writer = new System.IO.StreamWriter(textFilePath, true))
                                {
                                    writer.WriteLine(text);
                                    writer.Close();
                                }
                            }
                        }
                    }
                }
            }

            return 0;
        }

        #region PDFを画像変換

        /// <summary>
        /// BitmapをBitmapSourceに変換する
        /// </summary>
        /// <param name="bmp"></param>
        /// <returns></returns>
        private static System.Windows.Media.Imaging.BitmapSource ConvertToBitmapSource(Bitmap bmp)
        {
            using (var dest = new Bitmap(bmp.Width, bmp.Height))
            using (var g = Graphics.FromImage(dest))
            {
                g.DrawImage(bmp, 0, 0, bmp.Width, bmp.Height);
                var hbitmap = dest.GetHbitmap();
                BitmapSource bitmapSource = Imaging.CreateBitmapSourceFromHBitmap(hbitmap, IntPtr.Zero, Int32Rect.Empty,
                    BitmapSizeOptions.FromWidthAndHeight(bmp.Width, bmp.Height));
                DeleteObject(hbitmap);

                return bitmapSource;
            }
        }

        /// <summary>
        /// 画像を保存
        /// </summary>
        /// <param name="bitmapSource"></param>
        /// <param name="imageFilePath"></param>
        /// <param name="extension"></param>
        /// <returns></returns>
        private static bool SaveImage(BitmapSource bitmapSource, string imageFilePath, string extension = "png")
        {
            using (var outstream = new System.IO.FileStream(imageFilePath, System.IO.FileMode.Create))
            {
                var encoder = GetEncoder(extension);

                encoder.Frames.Add(BitmapFrame.Create(bitmapSource));
                encoder.Save(outstream);
                outstream.Flush();
            }

            return true;
        }

        /// <summary>
        /// 画像のエンコーダを取得する
        /// </summary>
        /// <param name="extension"></param>
        /// <returns></returns>
        private static BitmapEncoder GetEncoder(string extension)
        {
            switch (extension)
            {
                case "png":
                    return new PngBitmapEncoder();
                case "bmp":
                    return new BmpBitmapEncoder();
                case "jpg":
                    return new JpegBitmapEncoder();
                case "gif":
                    return new GifBitmapEncoder();
                case "tif":
                    return new TiffBitmapEncoder();
            }

            throw new ArgumentException();
        }

        #endregion

        #region 変換後画像ファイル

        /// <summary>
        /// 画像ファイル読み込み
        /// </summary>
        /// <param name="path"></param>
        private void LoadImages(string path)
        {
            var imageFiles = (from x in System.IO.Directory.GetFiles(path)
                              where IMAGE_EXTENTIONS.Any(ext => System.IO.Path.GetExtension(x) == ext)
                              select x);

            _images.Clear();
            foreach (var imageFile in imageFiles)
            {
                using (var stream = new System.IO.FileStream(imageFile, System.IO.FileMode.Open))
                {
                    var bmpimg = new BitmapImage();
                    bmpimg.BeginInit();
                    bmpimg.CacheOption = BitmapCacheOption.OnLoad;
                    bmpimg.StreamSource = stream;
                    bmpimg.EndInit();

                    stream.Close();

                    var image = new Image()
                    {
                        Source = bmpimg,
                        Margin = new Thickness(10)
                    };
                    this._images.Add(image);
                }
            }
        }

        /// <summary>
        /// 画像をウインドウサイズにあわせて再設定
        /// </summary>
        private void UpdateImages()
        {
            SPnlImages.Children.Clear();
            foreach (var image in _images)
            {
                image.MaxHeight = ScrMain.ViewportHeight - image.Margin.Top - image.Margin.Bottom;
                SPnlImages.Children.Add(image);
            }
            SPnlImages.UpdateLayout();
        }

        /// <summary>
        /// 読み込み済みの画像をクリア
        /// </summary>
        private void ClearImages()
        {
            SPnlImages.Children.Clear();
            this._images.Clear();
            DoEvents();
        }

        #endregion

        #region OCR

        /// <summary>
        /// BitmapSourceの内容をMicrosoft OCRで文字認識させる
        /// </summary>
        /// <param name="bmpsource"></param>
        /// <returns></returns>
        private static async Task<string> RecognizeFromBitmapAsync(OcrEngine engine, BitmapSource bmpsource)
        {
            var encoder = new BmpBitmapEncoder();
            encoder.Frames.Add(System.Windows.Media.Imaging.BitmapFrame.Create(bmpsource));

            using (var stream = new MemoryStream())
            {
                encoder.Save(stream);

                // BitmapSourceの内容をSoftwareBitmapに変換
                using (var softwareBitmap = await LoadImageFromByteArrayAsync(stream.ToArray()))
                {
                    var result = await engine.RecognizeAsync(softwareBitmap);
                    var lines = new List<string>();
                    result.Lines.ToList().ForEach(line => lines.Add(line.Text));

                    return string.Join("\n", lines);
                }
            }
        }

        /// <summary>
        /// byte型の配列からSoftwareBitmapを取得する
        /// </summary>
        /// <param name="ba"></param>
        /// <returns></returns>
        private static async Task<SoftwareBitmap> LoadImageFromByteArrayAsync(byte[] ba)
        {
            using (var imrastream = new InMemoryRandomAccessStream())
            using (var outstream = imrastream.GetOutputStreamAt(0))
            using (var writer = new DataWriter(outstream))
            {
                writer.WriteBytes(ba);
                await writer.StoreAsync();
                if (!await outstream.FlushAsync()) throw new InvalidOperationException();

                var decoder = await Windows.Graphics.Imaging.BitmapDecoder.CreateAsync(imrastream);

                return await decoder.GetSoftwareBitmapAsync(BitmapPixelFormat.Bgra8, BitmapAlphaMode.Premultiplied);
            }
        }

        #endregion

        #region ダイアログ

        /// <summary>
        /// ファイル選択ダイアログを開く
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="title"></param>
        /// <param name="current"></param>
        /// <returns></returns>
        private Tuple<bool, string> OpenFileDialog(string filter, string title, string current)
        {

            var dlg = new OpenFileDialog();
            dlg.FilterIndex = 1;
            dlg.Filter = filter;
            dlg.Title = title;
            dlg.CheckFileExists = true;
            dlg.CheckPathExists = true;
            dlg.ValidateNames = true;
            if (String.IsNullOrEmpty(current))
            {
                dlg.InitialDirectory = System.Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
            }
            else
            {
                var path = current.Trim();
                dlg.InitialDirectory = System.IO.Path.GetDirectoryName(current);
                if (System.IO.File.Exists(path))
                {
                    dlg.FileName = path;
                }
            }

            return dlg.ShowDialog() == true ? new Tuple<bool, string>(true, dlg.FileName) : new Tuple<bool, string>(false, "");
        }

        /// <summary>
        /// フォルダ選択ダイアログを開く
        /// </summary>
        /// <param name="title"></param>
        /// <param name="current"></param>
        /// <returns></returns>
        private Tuple<bool, string> OpenFolderDialog(string title, string current)
        {

            var dlg = new FolderBrowserDialog();
            dlg.Description = title;
            dlg.SelectedPath = current;
            if (String.IsNullOrEmpty(current))
            {
                dlg.SelectedPath = System.Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
            }
            else
            {
                dlg.SelectedPath = current;
            }

            return dlg.ShowDialog() == System.Windows.Forms.DialogResult.OK ? new Tuple<bool, string>(true, dlg.SelectedPath) : new Tuple<bool, string>(false, "");
        }

        #endregion

        /// <summary>
        /// 
        /// </summary>
        private static void DoEvents()
        {
            DispatcherFrame frame = new DispatcherFrame();
            var callback = new DispatcherOperationCallback((o) =>
            {
                ((DispatcherFrame)o).Continue = false;
                return null;
            });
            Dispatcher.CurrentDispatcher.BeginInvoke(DispatcherPriority.Background, callback, frame);
            Dispatcher.PushFrame(frame);
        }
    }
}
